@@ -8,10 +8,13 @@
  *
  * See the file "license.terms" for information on usage and redistribution of
  * this file, and for a DISCLAIMER OF ALL WARRANTIES.
- *
- * RCS: @(#) $Id: tclFCmd.c,v 1.43.2.1 2008/07/21 14:56:10 patthoyts Exp $
  */
 
+#ifndef _WIN64
+/* See [Bug 2935503]: file mtime sets wrong time */
+#   define _USE_32BIT_TIME_T
+#endif
+
 #include "tclInt.h"
 
 /*
@@ -528,7 +531,7 @@ CopyRenameOneFile(
 	 * 16 bits and we get collisions. See bug #2015723.
 	 */
 
-#ifndef WIN32
+#if !defined(WIN32) && !defined(__CYGWIN__)
 	if ((sourceStatBuf.st_ino != 0) && (targetStatBuf.st_ino != 0)) {
 	    if ((sourceStatBuf.st_ino == targetStatBuf.st_ino) &&
 		    (sourceStatBuf.st_dev == targetStatBuf.st_dev)) {
@@ -952,7 +955,7 @@ TclFileAttrsCmd(
     int result;
     CONST char ** attributeStrings;
     Tcl_Obj* objStrings = NULL;
-    int numObjStrings = -1;
+    int numObjStrings = -1, didAlloc = 0;
     Tcl_Obj *filePtr;
 
     if (objc < 3) {
@@ -985,9 +988,8 @@ TclFileAttrsCmd(
 		Tcl_AppendResult(interp, "could not read \"",
 			TclGetString(filePtr), "\": ", Tcl_PosixError(interp),
 			NULL);
-		return TCL_ERROR;
 	    }
-	    goto end;
+	    return TCL_ERROR;
 	}
 
 	/*
@@ -1005,12 +1007,16 @@ TclFileAttrsCmd(
 	}
 	attributeStrings = (CONST char **) TclStackAlloc(interp,
 		(1+numObjStrings) * sizeof(char*));
+	didAlloc = 1;
 	for (index = 0; index < numObjStrings; index++) {
 	    Tcl_ListObjIndex(interp, objStrings, index, &objPtr);
 	    attributeStrings[index] = TclGetString(objPtr);
 	}
 	attributeStrings[index] = NULL;
+    } else if (objStrings != NULL) {
+	Tcl_Panic("must not update objPtrRef's variable and return non-NULL");
     }
+
     if (objc == 0) {
 	/*
 	 * Get all attributes.
@@ -1071,6 +1077,10 @@ TclFileAttrsCmd(
 		"option", 0, &index) != TCL_OK) {
 	    goto end;
 	}
+	if (didAlloc) {
+	    TclFreeIntRep(objv[0]);
+	    objv[0]->typePtr = NULL;
+	}
 	if (Tcl_FSFileAttrsGet(interp, index, filePtr,
 		&objPtr) != TCL_OK) {
 	    goto end;
@@ -1095,6 +1105,10 @@ TclFileAttrsCmd(
 		    "option", 0, &index) != TCL_OK) {
 		goto end;
     	    }
+	    if (didAlloc) {
+		TclFreeIntRep(objv[i]);
+		objv[i]->typePtr = NULL;
+	    }
 	    if (i + 1 == objc) {
 		Tcl_AppendResult(interp, "value for \"",
 			TclGetString(objv[i]), "\" missing", NULL);
@@ -1109,20 +1123,20 @@ TclFileAttrsCmd(
     result = TCL_OK;
 
   end:
-    if (numObjStrings != -1) {
+    if (didAlloc) {
 	/*
 	 * Free up the array we allocated.
 	 */
 
 	TclStackFree(interp, (void *)attributeStrings);
+    }
 
+    if (objStrings != NULL) {
 	/*
 	 * We don't need this object that was passed to us any more.
 	 */
 
-	if (objStrings != NULL) {
-	    Tcl_DecrRefCount(objStrings);
-	}
+	Tcl_DecrRefCount(objStrings);
     }
     return result;
 }
