@@ -260,7 +260,7 @@ TclpThreadExit(
 Tcl_ThreadId
 Tcl_GetCurrentThread(void)
 {
-    return (Tcl_ThreadId) GetCurrentThreadId();
+    return (Tcl_ThreadId) INT2PTR(GetCurrentThreadId());
 }
 
 /*
